#include "defines.h"
#include "sgemv_col_major_kernels.cu"
#include <cublas.h>

#if not defined (_TESLA) && not defined (_FERMI)
#error Requires an architechture flag (either _TESLA or _FERMI) to compile.
#endif

#define switchParameters alpha, A_d, lda, x_d, y_d
#include "level2_templatemacros.h"

#ifdef _ATUNE

// When building atune
#define arguments sgemv_parameters
#define call_arguments sgemv_call_parameters
#define switchMacro1 workSizeSwitch_mn
#define switchMacro2 workSizeSwitchLarge_mn
#define templatedec
#include "magma.h"

#else

// When building library
#define BLOCKSIZE BLOCKSIZE_m * BLOCKSIZE_n
#define blockSize_m BLOCKSIZE_m
#define workSize_m WORKSIZE_m
#define workSize_n WORKSIZE_n
#define arguments sgemv_arguments
#define call_arguments sgemv_call_arguments
#define switchMacro1 call_kernel
#define switchMacro2 call_kernel
#define templatedec template<int BLOCKSIZE_m, int BLOCKSIZE_n, int WORKSIZE_m, int WORKSIZE_n>

#endif

templatedec
static int sgemv_col_major_v1(arguments)
{		
	//
	// One-thread-per-row version
	//

	// Clear output vector if beta == 0.0
	if (beta == 0.0) cudaAssert( cudaMemset(y_d, 0, m * sizeof(float)) );

	// Determine size of grid
	int gridSize_m = (m + blockSize_m * workSize_m - 1) / (blockSize_m * workSize_m);
	int gridSize_n = 1;
	
	// Sizes of the tiled part of A
	int mtile = (m / (blockSize_m * workSize_m)) * (blockSize_m * workSize_m);
	int ntile = (n / blockSize_m) * blockSize_m;
		
	// Kernel specifications
	dim3 dimGrid(gridSize_m, gridSize_n);
	dim3 dimBlock(blockSize_m);
		
	// Call nested switch macros for all templated values
	//printf("\nOne-thread-per-row-%d: m=%d n=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d work_m=%d work_n=%d\n", version, m, n, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, workSize_m, workSize_n);
	switchMacro1(sgemv_one_thread_per_row_kernel, BLOCKSIZE, workSize_m, workSize_n);
	return 0;
}

templatedec
static int sgemv_col_major_v2(arguments)
{		
	//
	// Generic version
	//
		
	// Clear output vector if beta == 0.0
	if (beta == 0.0) cudaAssert( cudaMemset(y_d, 0, m * sizeof(float)) );

	// Determine size of grid
	int gridSize_m = (m + blockSize_m * workSize_m - 1) / (blockSize_m * workSize_m);
	int gridSize_n = (n + blockSize_m * workSize_n - 1) / (blockSize_m * workSize_n);
		
	// Sizes of the tiled part of A
	int mtile = (m / (blockSize_m * workSize_m)) * (blockSize_m * workSize_m);
	int ntile = (n / (blockSize_m * workSize_n)) * (blockSize_m * workSize_n);
			
	// Kernel specifications
	dim3 dimGrid(gridSize_m, gridSize_n);
	dim3 dimBlock(blockSize_m);
		
	// Call nested switch macros for all templated values
	//printf("\nGeneric-%d: m=%d n=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d work_m=%d work_n=%d\n", version, m, n, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, workSize_m, workSize_n);
	switchMacro1(sgemv_generic_kernel, BLOCKSIZE, workSize_m, workSize_n);
	return 0;
}

templatedec
static int sgemv_col_major_v3(arguments)
{		
	//
	// several-columns version
	//

	// Clear output vector if beta == 0.0
	if (beta == 0.0) cudaAssert( cudaMemset(y_d, 0, m * sizeof(float)) );

#ifndef workSize_m
	// Adjust columns per thread (maximum allowed is currently 16)
	if (workSize_m == 1)
	{
		workSize_m = blockSize_m / m;
		if (workSize_m > 16) workSize_m = 16;
	}
#endif

	// Determine size of grid
	int gridSize_m = 1;
	int gridSize_n = (n + blockSize_m * workSize_m * workSize_n - 1) / (blockSize_m * workSize_m * workSize_n);

	// Sizes of the tiled part of A
	int mtile = 0;
	int ntile = (n / (blockSize_m * workSize_m * workSize_n)) * (blockSize_m * workSize_m * workSize_n);

	// Kernel specifications
	dim3 dimGrid(gridSize_m, gridSize_n);
	dim3 dimBlock(blockSize_m);
			
	// Call nested switch macros for all templated values
	//printf("\nSeveral-cols-%d: m=%d n=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d work_m=%d work_n=%d\n", version, m, n, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, workSize_m, workSize_n);
	switchMacro2(sgemv_several_cols_kernel, BLOCKSIZE, workSize_m, workSize_n);
	return 0;
}

/*
// Only for PPAM paper!!

templatedec
static int sgemv_col_major_v4(arguments)
{		
	//
	// Several-threads-per-row version
	//
		
	// Clear output vector (assume beta=0.0)
	cudaAssert( cudaMemset(y_d, 0, m * sizeof(float)) );
		
	// Determine size of grid
	int gridSize_m = (m + blockSize_m - 1) / blockSize_m;
	int gridSize_n = (n + blockSize_m * workSize_n - 1) / (blockSize_m * workSize_n);
		
	// Sizes of the tiled part of A
	int mtile = (m / blockSize_m) * blockSize_m;
	int ntile = (n / (blockSize_m * workSize_n)) * (blockSize_m * workSize_n);
			
	// Kernel specifications
	dim3 dimGrid(gridSize_m, gridSize_n);
	dim3 dimBlock(blockSize_m);
		
	// Call nested switch macros for all templated values
	//printf("\nSeveral_threads_per_row-%d: m=%d n=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d work_m=%d work_n=%d\n", version, m, n, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, workSize_m, workSize_n);
	switchMacro1(sgemv_several_threads_per_row_kernel, BLOCKSIZE, workSize_m, workSize_n);
	return 0;
}
*/

#ifdef _ATUNE

// When building atune
template <int BLOCKSIZE_n>
int sgemv_col_major(sgemv_parameters);

template <>
int sgemv_col_major<BLOCKSIZE>(sgemv_parameters)
{
	// skip sizes when we know this version is inefficient on Fermi
#ifdef _FLOAT
	if (version == 1 && (m > 0.5 * (1536 * 14) || (n <= 2048 && workSize_m == 1)))
#else
	if (version == 1 && (m > 0.2 * (1536 * 14) || (n <= 2048 && workSize_m == 1)))
#endif
	{
		return sgemv_col_major_v1(sgemv_call_parameters);
	}
	else if (version == 2)
	{
		return sgemv_col_major_v2(sgemv_call_parameters);
	}
	// use only this version for short and fat sizes
	else if (version == 3 && (2 * m <= blockSize_m && workSize_m * m <= blockSize_m && n > blockSize_m))
	{
		return sgemv_col_major_v3(sgemv_call_parameters);
	}
	/*
	else if (version == 4)
	{
		return sgemv_col_major_v4(sgemv_call_parameters);
	}
	*/

	// return "not run" signal to auto-tuner
	return -1;
}

#else

//When building library
#include "atune_headers/sgemv_col_major.h"
int atune_sgemv_row_major(sgemv_arguments);

int glas_sgemv(char trans, int m, int n, float alpha, float *A_d, int lda, float *x_d, int incx, float beta, float *y_d, int incy)
{
	if ((beta == 0.0 || beta == 1.0) && incx == 1 && incy == 1)
	{
		if (trans == 'N' || trans == 'n')
		{
			return atune_sgemv_col_major(sgemv_call_arguments);
		}
		else if (trans == 'T' || trans == 't')
		{
			return atune_sgemv_row_major(sgemv_call_arguments);
		}
		else 
		{
			printf("Call to glas_sgemv: parameter 1 has an illegal value!\n"); exit(EXIT_FAILURE);
		}
	}
	else
	{
		printf("Call to sgemv: parameter 8, 9 or 11 has an unsupported value (%d, %f, %d)!\n", incx, beta, incy); exit(EXIT_FAILURE);
	}
	return -1;
}

#endif
