#include "defines.h"
#include "dense.h"
#include "glas.h"
#include "strsv_col_major_templatemacros.h"
#include "strsv_col_major_kernels.cu"
//#include "magma.h"

#if not defined (_TESLA) && not defined (_FERMI)
#error Requires an architechture flag (either _TESLA or _FERMI) to compile.
#endif

#ifdef _ATUNE

// When building atune
#define arguments strsv_parameters
#define call_arguments strsv_call_parameters
#define switchMacro1 workSizeSwitch1Pow_m_Pow_n
#define switchMacro2 workSizeSwitch2Pow_m_Pow_n
#define templatedec

#else

// When building library
#define BLOCKSIZE_m BLOCKSIZE
#define blockSize_m BLOCKSIZE_m
#define blockSize_n BLOCKSIZE_n
#define workSize_m WORKSIZE_m
#define workSize_n WORKSIZE_n
#define arguments strsv_arguments
#define call_arguments strsv_call_arguments
#define switchMacro1 call_kernel1
#define switchMacro2 call_kernel2
#define templatedec template<int BLOCKSIZE_m, int BLOCKSIZE_n, int WORKSIZE_m, int WORKSIZE_n>
#include <cublas.h>

#endif

templatedec
static int strsv_col_major_v1(arguments)
{		
		//
		// Generic version 1 (Sequential design)
		//

		// Determine size of grid
		int gridSize_m = (m + blockSize_m * workSize_m - 1) / (blockSize_m * workSize_m);
		int gridSize_n = 1;
		
		// Sizes of the tiled part of A
		int mtile = (m / (blockSize_m * workSize_m)) * (blockSize_m * workSize_m);
		int ntile = 1;
			
		// Kernel specifications
		dim3 dimGrid(gridSize_m, gridSize_n);
		dim3 dimBlock(blockSize_m);

		// Call kernels in correct order
		//printf("\nGeneric-%d: m=%d lda=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d work_m=%d work_n=%d\n", version, m, lda, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, workSize_m, workSize_n);
		strsv_generic_kernel1_solve<BLOCKSIZE, 1, 1><<< dim3(1), dimBlock >>>(A_d, lda, x_d, work_d, m, mtile, ntile);
		for (int i = 1; i < gridSize_m; ++i)
		{
			strsv_generic_kernel1_column_rest<BLOCKSIZE, 1, 1><<< dim3(gridSize_m - i), dimBlock >>>(A_d + i * BLOCKSIZE + (i - 1) * lda * BLOCKSIZE, lda, x_d + i * BLOCKSIZE, work_d, m - i * BLOCKSIZE, mtile, ntile);
		}
		return 0;
}

templatedec
static int strsv_col_major_v2(arguments)
{		
		//
		// Generic version 2 (Parallel design)
		//
		// - Requires workspace >= min(tileSize, m) * min(tileSize, m) + m
		//

		int tileSize = blockSize_m * workSize_m;
		int tiles = ((m + tileSize - 1)/ tileSize);

		// Determine size of grid
		int gridSize_m = (m + blockSize_m - 1) / blockSize_m;
		int gridSize_n = 1;
		
		// Sizes of the tiled part of A
		int mtile = (m / tileSize) * tileSize;
		int ntile = 1;
			
		// Kernel specifications
		dim3 dimGrid(gridSize_m, gridSize_n);
		dim3 dimBlock(blockSize_m, blockSize_n);

		// Maybe workspace is provided in work_d, otherwise allocate workspace
		float *xtemp_d = work_d, *Ainv_d = work_d + m;
		if (work_d == NULL)
		{
			printf("VI ALLOKERER HUKOMMELSE\n");
			cudaAssert(cudaMalloc(&xtemp_d, m + gridSize_m * tileSize * tileSize * sizeof(Ainv_d[0])));
			Ainv_d = xtemp_d + m;
		}
//#define _DEBUG
#ifdef _DEBUG
		cudaMemset(Ainv_d, 0, tiles * tileSize * tileSize * sizeof(float));
		printf("\nGeneric-%d: m=%d lda=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d block_n=%d work_m=%d work_n=%d\n", 2, m, lda, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, blockSize_n, workSize_m, workSize_n);
#endif
		// Call nested switch macros for all templated values
		switchMacro1(strsv_inv_diag_shared, workSize_m, blockSize_m);
		// Expand diagonal inverses to desired outer block size (blockSize_m * workSize_m)
		for (int size = blockSize_m; size < tileSize; size *= 2)
		{
			int size2 = 2 * size;
			int gridSize_m = (m + size2 - 1) / size2;
			int gridSize_n = size / workSize_n;
			int mtile = (m / size2) * size2;
			int ntile = (m / (2 * workSize_n)) * (2 * workSize_n);
			dim3 dimGrid(gridSize_m, gridSize_n);
			dim3 dimBlock(size);
#ifdef _DEBUG				
			printf("\nExpand-%d: size=%d tileSize=%d tiles=%d m=%d lda=%d grid_m=%d grid_n=%d mtile=%d ntile=%d block_m=%d block_n=%d work_m=%d work_n=%d\n", 2, size, tileSize, tiles, m, lda, gridSize_m, gridSize_n, mtile, ntile, blockSize_m, blockSize_n, workSize_m, workSize_n);
#endif
			switchMacro2(strsv_inv_diag_expand, size, workSize_n);
		}
#ifdef _DEBUG				
		dense::matrix A, Ainv;
		A.zeroset(m, m);
		Ainv.zeroset(tileSize, tileSize);
		cudaAssert(cudaMemcpy(A.data, A_d, sizeof(A_d[0]) * A.m * A.n, D2H));
		A.print("A");
		for (int i = 0; i < tiles; ++i)
		{
			cudaAssert(cudaMemcpy(Ainv.data, Ainv_d + tileSize * tileSize * i, sizeof(A_d[0]) * tileSize * tileSize, D2H));
			Ainv.print("Ainv");
		}
#endif
		for (int i = 0; i < m; i += tileSize)
		{
			int size = min(tileSize, m - i);
			//cublasSgemm('N', 'N', size, 1, size, 1.0, Ainv_d + i * tileSize, tileSize, x_d + i, m, 0, xtemp_d + i, m);  
			glas_sgemv('N', size, size, 1.0, Ainv_d + i * tileSize, tileSize, x_d + i, 1, 0.0, xtemp_d + i, 1);
			if (i + tileSize >= m) break;
			//cublasSgemm('N', 'N', m - i - tileSize, 1, tileSize, -1.0, A_d + tileSize + i + i * lda, lda, xtemp_d + i, m, 1.0, x_d + i + tileSize, m);  
			glas_sgemv('N', m - i - tileSize, tileSize, -1.0, A_d + tileSize + i + i * lda, lda, xtemp_d + i, 1, 1.0, x_d + i + tileSize, 1);
		}
		glas_scopy(m, xtemp_d, 1, x_d, 1);
		if (work_d == NULL) 
			cudaFree(xtemp_d);
		return 0;
}

#ifdef _ATUNE

// When building atune
template <int BLOCKSIZE_n>
int strsv_col_major(strsv_parameters);

template <>
int strsv_col_major<BLOCKSIZE>(strsv_parameters)
{
	// Skip sizes when we know this version is inefficient
	if (version == 1 && (m > 1024))
	{
		return strsv_col_major_v1(strsv_call_parameters);
	}
	// Works only for sizes of power two
#if defined (BLOCKSIZE) && (BLOCKSIZE == 16 || BLOCKSIZE == 32 || BLOCKSIZE == 64 || BLOCKSIZE == 128 || BLOCKSIZE == 256 || BLOCKSIZE == 512)
	else if (version == 2 && blockSize_m * workSize_m <= m)
	{
		return strsv_col_major_v2(strsv_call_parameters);
	}
#endif

	// return "not run" signal to auto-tuner
	return -1;
}

#else

//When building library
#include "atune_headers/strsv_col_major.h"
int glas_strsv(char uplo, char trans, char diag, int m, float *A_d, int lda, float *x_d, int incx, float *work_d)
{
	if (uplo == 'L' && diag == 'N' && incx == 1)
	{
		if (trans == 'N' || trans == 'n')
		{
			return atune_strsv_col_major(strsv_call_arguments);
		}
		else if (trans == 'T' || trans == 't')
		{
			//return atune_strsv_row_major(strsv_call_arguments);
		}
		else 
		{
			printf("Call to glas_strsv: parameter 2 has an illegal value!\n"); exit(EXIT_FAILURE);
		}
	}
	else
	{
		printf("Call to strsv: parameter 1, 3 or 8 has an unsupported value (%s, %s, %d)!\n", uplo, diag, incx); exit(EXIT_FAILURE);
	}
	return -1;
}

#endif

/*
#if BLOCKSIZE == 64
#define NB 64
#include "magma_strsm.cu"
#endif

#if BLOCKSIZE == 64
	else if (version == 3 && m >= NB)
	{
		//
		// Magma version
		//
		float *Ainv_d, *xtemp_d;
		int gridSize_m = m/NB;
		int blockSize_m = NB;
		Ainv_d = y_d + m; xtemp_d = y_d;
		
		//cudaMalloc((void**)&Ainv_d, gridSize_m * blockSize_m * blockSize_m * sizeof(float));
		//cudaMalloc((void**)&xtemp_d, m * sizeof(float));
		cudaMemset(xtemp_d, 0, m * sizeof(float));
		cudaMemset(Ainv_d, 0, gridSize_m * blockSize_m * blockSize_m * sizeof(float));
		
		diag_strtri(m, 'L', 'N', A_d, Ainv_d, lda);
		int MM = min (NB, m); 
		int ldb = m;
		float *b = x_d, *d_x = xtemp_d;
		int N = 1, M = m;
		cublasSgemm ('N', 'N', MM, 1, MM, 1.0, Ainv_d, NB, b, ldb, 0, xtemp_d, m);  
		if (NB<m)
		{
			
			cublasSgemm ('N', 'N', m-NB, 1, NB, -1.0, A_d+NB, lda, xtemp_d, m, 1.0, b+NB, ldb);
			
			for (int i=NB; i<m; i+=NB)
			{
				MM = min (m-i, NB);
				cublasSgemm ('N', 'N', MM, 1, MM, 1.0, Ainv_d+i*NB, NB, b+i, ldb, 0, xtemp_d+i, m);  
				
				if (i+NB>=m)
					break;
				
				cublasSgemm ('N', 'N', m-i-NB, 1, NB, -1.0, A_d+i*lda+i+NB, lda, xtemp_d+i, m, 1.0, b+i+NB, ldb);
			}
		}

		b_copy();
		//cudaFree(Ainv_d);
		//cudaFree(xtemp_d);
		return 0;
	}
#endif			
#endif

	// return "not run" signal to auto-tuner
	return -1;
}
*/
