#include "atomicAdd.h"
#include "strsv_col_major_kernelmacros.h"

// 
// Generic matrix-vector multiplication kernel.
//
// - Designed for architectures with atomicCAS, i.e., sm_11 and higher.
// - Split into 4 separate device kernels.
// -----------------
// |         |     |
// |    1    |  2  |
// |         |     |
// |---------------|
// |    3    |  4  |
// -----------------
// - Valid for all cases.
// - Uses atomicCAS (sm_11-sm_13) or atomicAdd (>=sm_20) for grid level reduction.
// - Unrolling of inner loops
//     * Given by a constant: full unroll
//     * Given by a variable: unroll level 4
//     1) to reduce instruction count from loop
//     2) to collect shared loads unsigned into 128 bit transactions (compiler should do that)
//


/*
template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m, MIN_BLOCKS_PER_MP)
__global__ void strsv_inv_diag_expand_step1(const float *A, int lda, const float *B, int ldb, float *C, int tiles, int m, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int row = 2 * blockIdx.x * BLOCKSIZE_m;
	int col = blockIdx.y * WORKSIZE_n;
	int jmpa = row * (1 + lda);
	int jmpb = row * ldb + (blockIdx.x & (tiles - 1)) * (2 * BLOCKSIZE_m);
	__shared__ float B_s[BLOCKSIZE_m * WORKSIZE_n];
	A += tid + jmpa;
	B += tid + col * ldb + jmpb;
	C += tid + col * ldb + jmpb;

	// Copy columns of B into shared memory
	for (int i = 0; i < WORKSIZE_n; ++i)
	{
		B_s[tid + i * BLOCKSIZE_m] = B[i * ldb];
	}
	__syncthreads();

	// Do dot product with row of A and columns of B
	initialize_sums(WORKSIZE_n);
    #pragma unroll
	for (int k = 0; k < BLOCKSIZE_m; ++k)
	{
		add_to_sums(WORKSIZE_n, BLOCKSIZE_m); A += lda;
	}
	write_sums(WORKSIZE_n, ldb);
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m, MIN_BLOCKS_PER_MP)
__global__ void strsv_inv_diag_expand_step2(const float *A, int lda, float *B, int ldb, float *C, int tiles, int m, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int row = 2 * blockIdx.x * BLOCKSIZE_m;
	int col = blockIdx.y * WORKSIZE_n;
	int jmp = row * ldb + (blockIdx.x & (tiles - 1)) * (2 * BLOCKSIZE_m);
	__shared__ float B_s[BLOCKSIZE_m * WORKSIZE_n];
	A += tid + jmp;
	B += tid + col * ldb + jmp;
	C += tid + col * ldb + jmp;

	// Copy columns of B into shared memory
	for (int i = 0; i < WORKSIZE_n; ++i)
	{
		B_s[tid + i * BLOCKSIZE_m] = B[i * ldb];
		// Erase values above diagonal used as workspace in step 1
		B[i * ldb] = 0.0f;
	}
	__syncthreads();

	// Do dot product with row of A and columns of B
	initialize_sums(WORKSIZE_n);
    #pragma unroll
	for (int k = 0; k < BLOCKSIZE_m; ++k)
	{
		// NB: minus in step 2
		sub_to_sums(WORKSIZE_n, BLOCKSIZE_m); A += ldb;
	}
	write_sums(WORKSIZE_n, ldb);
}
*/


/*

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void kernel1(const float* A, const float* x, float* y, float *x_s, int m, int tid)
{
	//
	// The top-left fully tiled part of the Matrix
	//
	initialize_sums(WORKSIZE_m);
	for (int j = 0; j < BLOCKSIZE_m * WORKSIZE_n; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
        #pragma unroll
		inner_loop(BLOCKSIZE_m);
	}
	atomicAdd_sums(BLOCKSIZE_m);
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void kernel2(const float* A, const float* x, float* y, float *x_s, int m, int nrest, int tid)
{
	//
	// The top-right m-tiled part of the Matrix
	//
	int ntile = (nrest / BLOCKSIZE_m) * BLOCKSIZE_m;
	initialize_sums(WORKSIZE_m);
	for (int j = 0; j < ntile; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
        #pragma unroll
		inner_loop(BLOCKSIZE_m);
	}
	__syncthreads();
	if (tid < nrest - ntile)
		x_s[tid]  = x[ntile];
	__syncthreads();
    #pragma unroll 4
	inner_loop(nrest - ntile);
	atomicAdd_sums(BLOCKSIZE_m);
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void kernel3(const float* A, const float* x, float* y, float *x_s, int m, int mrest, int tid)
{
	//
	// The bottom-left n-tiled part of the Matrix
	//
	initialize_sums(WORKSIZE_m);
	for (int j = 0; j < BLOCKSIZE_m * WORKSIZE_n; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
		if (tid < mrest)
		{
            #pragma unroll
			inner_loop_check_bounds(tid, mrest, BLOCKSIZE_m);
		}
	}
	if (tid < mrest)
	{
		atomicAdd_sums_check_bounds(tid, mrest, BLOCKSIZE_m);
	}
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void kernel4(const float* A, const float* x, float* y, float *x_s, int m, int mrest, int nrest, int tid)
{
	//
	// The bottom-right rest of the Matrix
	//
	int ntile = (nrest / BLOCKSIZE_m) * BLOCKSIZE_m;
	initialize_sums(WORKSIZE_m);
	for (int j = 0; j < ntile; j += BLOCKSIZE_m)
	{
		__syncthreads();
		x_s[tid]  = x[j];
		__syncthreads();
		if (tid < mrest)
		{
            #pragma unroll
			inner_loop_check_bounds(tid, mrest, BLOCKSIZE_m);
		}
	}
	__syncthreads();
	if (tid < nrest - ntile)
		x_s[tid]  = x[ntile];
	__syncthreads();
	if (tid < mrest)
	{
        #pragma unroll 4
		inner_loop_check_bounds(tid, mrest, nrest - ntile);
		atomicAdd_sums_check_bounds(tid, mrest, BLOCKSIZE_m);
	}
}

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m, MIN_BLOCKS_PER_MP)
__global__ void strsv_generic_kernel(const float* A, const float* x, float* y, int m, int n, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int row = blockIdx.x * BLOCKSIZE_m * WORKSIZE_m;
	int col = blockIdx.y * BLOCKSIZE_m * WORKSIZE_n;
	__shared__ float x_s[BLOCKSIZE_m];
	A += tid + row + col * m;
	x += tid + col;
	y += tid + row;
	if (row !=  mtile)
	{
		if (col != ntile)
		{	
			kernel1<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(A, x, y, x_s, m, tid);
		}
		else
		{
			kernel2<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(A, x, y, x_s, m, n - ntile, tid);
		}
	}
	else
	{
		if (col != ntile)
		{
			kernel3<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(A, x, y, x_s, m, m - mtile, tid);
		}
		else
		{
			kernel4<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(A, x, y, x_s, m, m - mtile, n - ntile, tid);
		}
	}
}
//
// Kernel that works for one block only
// -----------------------------------------
// Threads    : BLOCKSIZE_m x WORKSIZE_n
// Shared mem : BLOCKSIZE_m x BLOCKSIZE_m + BLOCKSIZE_m
//
template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m * WORKSIZE_n, MIN_BLOCKS_PER_MP)
__global__ void strsv_generic_kernel5(const float* A, float* x, float* y, int m, int lda, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int row = blockIdx.x * BLOCKSIZE_m * WORKSIZE_m;
	//int col = blockIdx.y;// * BLOCKSIZE_m * WORKSIZE_n;
	int col = threadIdx.y;
	__shared__ float xsmem[BLOCKSIZE_m];
	__shared__ float Asmem[BLOCKSIZE_m * BLOCKSIZE_m];
	A += tid + row + col * lda;
	x += tid + row;
	float res;
	float *x_s = xsmem;
	if (tid < BLOCKSIZE_m) { res = x[0]; x_s[tid] = 0; }
	tid += col * BLOCKSIZE_m;
	float *A_s = Asmem;
	#pragma unroll
	for (int j = 0; j < BLOCKSIZE_m; j += WORKSIZE_n)
	{
		A_s[tid + j * lda] = A[j * lda];
	}
	__syncthreads();
	if (tid < BLOCKSIZE_m) {
        #pragma unroll
		for (int i = 0; i < BLOCKSIZE_m; ++i)
		{
			if (tid == i) x_s[0] = res / A_s[tid];
			__syncthreads();
			res -= A_s[tid] * x_s[0];
			++x_s;
			A_s += BLOCKSIZE_m;
		}
	}
	// set output
	if (tid < BLOCKSIZE_m) x[0] = xsmem[tid];
}

// Kernel that works for one block only
// -----------------------------------------
// Threads    : BLOCKSIZE_m
// Shared mem : BLOCKSIZE_m
//
template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m, MIN_BLOCKS_PER_MP)
__global__ void strsv_generic_kernel2(const float* A, float* x, float* y, int m, int lda, int mtile, int ntile)
{
	__shared__ volatile float x_s[BLOCKSIZE_m];
	int tid = threadIdx.x;
	if (tid < m)
	{
		A += tid;
		x += tid;
		float res = x[0];
		block_trsv<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(A, x_s, res, lda, tid);
		x[0] = x_s[tid];
	}
}

//
// Kernel that works for multiple blocks
// -----------------------------------------
// Threads    : BLOCKSIZE_m
// Shared mem : BLOCKSIZE_m
// Device mem : 1
//
 __device__ volatile unsigned int counter = 0; 

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m, MIN_BLOCKS_PER_MP)
__global__ void strsv_generic_kernel1(const float* A, float* x, float* y, int m, int lda, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int row = blockIdx.x * BLOCKSIZE_m * WORKSIZE_m;
	__shared__ float x_s[BLOCKSIZE_m];
	float res;
	if (tid == 0 && blockIdx.x == 0) counter = 0;
	// read in right-hand-side of this block-row
	A += tid + row;
	x += tid + row;
	if (tid + row < m) res = x[0];
	for (int b = 0; b < blockIdx.x; ++b)
	{
		// wait for top block of block-column to finish
		volatile unsigned int value = counter;
		while (value <= b) { value = counter; };
		// all threads read in solution of top block-column
		x_s[tid] = x[b * BLOCKSIZE_m - row];
		__syncthreads();
		if (tid + row < m)
		{
			// do matrix-vector update of right-hand-side
            #pragma unroll
			block_gemv_loop(A, x_s, res, lda);
		}
	}
	if (tid + row < m)
	{
		// solve for x of this block-row
		block_trsv<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(A, x_s, res, lda, tid);
		//store result
		x[0] = x_s[tid];
	}
	__threadfence();
	if (tid == 0) 
	{
		// reset counter if this is the last block, otherwise signal block-row is done
		if (blockIdx.x == gridDim.x - 1) counter = 0;
		else counter++;
	}
}

*/

//
// Single Block Column-Sequential TRSV
//
// - Perform triangular solve A x_s = [res] of size BLOCKSIZE_m x BLOCKSIZE_m
//   using 1 block of BLOCKSIZE_m threads
//

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__device__ void block_strsv(const float *A, volatile float *x_s, float res, int lda, int tid)
{
    #pragma unroll
	for (int i = 0; i < BLOCKSIZE_m; i++)
	{
		if (tid == i) x_s[i] = res / A[0];
		__syncthreads();
		if (tid > i) {
			res -= A[0] * x_s[i]; A += lda;
		}
	}
}

//
// Single Block GEMV
//
// - Perform matrix-vector multiplication [res] = -A x_s of size BLOCKSIZE_m x BLOCKSIZE_m
//   using 1 block of BLOCKSIZE_m threads
//

#define block_sgemv_loop(A, x_s, res, lda)		\
	{											\
		for (int i = 0; i < BLOCKSIZE_m; i++)	\
		{										\
			res -= A[0] * x_s[i]; A += lda;		\
		}										\
	}


// Kernel that works for one block only
// -----------------------------------------
// Threads    : BLOCKSIZE_m
// Shared mem : BLOCKSIZE_m
//
// Flops(strsv) = n + n = n(n+1)!
//

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m, MIN_BLOCKS_PER_MP)
__global__ void strsv_generic_kernel1_solve(const float* A, int lda, float* x, float* y, int m, int mtile, int ntile)
{
	__shared__ volatile float x_s[BLOCKSIZE_m];
	int tid = threadIdx.x;
	if (tid < m)
	{
		A += tid;
		x += tid;
		float res = x[0];
		block_strsv<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(A, x_s, res, lda, tid);
		x[0] = x_s[tid];
	}
}

//
// Kernel that must be called repeatedly in correct order
// -------------------------------------------------
// Threads    : BLOCKSIZE_m
// Shared mem : BLOCKSIZE_m
//

template <int BLOCKSIZE_m, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_m, MIN_BLOCKS_PER_MP)
__global__ void strsv_generic_kernel1_column_rest(const float* A, int lda, float* x, float* y, int m, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int row = blockIdx.x * BLOCKSIZE_m * WORKSIZE_m;
	__shared__ float x_s[BLOCKSIZE_m];
	float res;
	// all threads read in solution of top block-column
	x_s[tid] = x[tid - BLOCKSIZE_m];
	__syncthreads();
	// read in right-hand-side of this block-row
	A += tid + row;
	x += tid + row;
	if (tid + row < m) 
	{
		res = x[0];
		// do matrix-vector update of right-hand-side
        #pragma unroll
		block_sgemv_loop(A, x_s, res, lda);
		if (blockIdx.x == 0)
		{
			// solve for x of this block-row
			block_strsv<BLOCKSIZE_m, WORKSIZE_m, WORKSIZE_n>(A, x_s, res, lda, tid);
			//store result
			x[0] = x_s[tid];
		}
		else
		{
			x[0] = res;
		}
	}
}

//
// Single Block SGEMM
//
// - Perform matrix-matrix C = A * B for shared memory blocks 
//   of size TILE x TILE, where TILE > 1
//

template <int TILE, int BLOCKSIZE_m, int BLOCKSIZE_n, bool MINUS>
__forceinline__ __device__ void block_sgemm2(volatile float *A1_s, volatile float *B1_s, volatile float *C_s, int tidx, int tid)
{
	float c[TILE / (2 * BLOCKSIZE_n)];
	if (tid < BLOCKSIZE_m * BLOCKSIZE_n)
	{
        #pragma unroll
		for (int i = 0; i < TILE / (2 * BLOCKSIZE_n); ++i)
			c[i] = 0;
        #pragma unroll
		for (int i = 0; i < TILE; ++i)
		{
			if (TILE >= (2 * BLOCKSIZE_n)) 
			{
				c[0] += A1_s[i * BLOCKSIZE_m] * B1_s[i];
			}
			if (TILE >= (4 * BLOCKSIZE_n)) 
			{
				c[1] += A1_s[i * BLOCKSIZE_m] * B1_s[i + 2 * BLOCKSIZE_m * BLOCKSIZE_n];
			}
			if (TILE >= (8 * BLOCKSIZE_n)) 
			{
				c[2] += A1_s[i * BLOCKSIZE_m] * B1_s[i + 4 * BLOCKSIZE_m * BLOCKSIZE_n];
				c[3] += A1_s[i * BLOCKSIZE_m] * B1_s[i + 6 * BLOCKSIZE_m * BLOCKSIZE_n];
			}
			if (TILE >= (16 * BLOCKSIZE_n)) 
			{
				c[4] += A1_s[i * BLOCKSIZE_m] * B1_s[i + 8 * BLOCKSIZE_m * BLOCKSIZE_n];
				c[5] += A1_s[i * BLOCKSIZE_m] * B1_s[i + 10 * BLOCKSIZE_m * BLOCKSIZE_n];
				c[6] += A1_s[i * BLOCKSIZE_m] * B1_s[i + 12 * BLOCKSIZE_m * BLOCKSIZE_n];
				c[7] += A1_s[i * BLOCKSIZE_m] * B1_s[i + 14 * BLOCKSIZE_m * BLOCKSIZE_n];
			}
		}
	}
	__syncthreads();
	if (tid < BLOCKSIZE_m * BLOCKSIZE_n)
	{
		if (!MINUS)
		{
			if (TILE >= (2 * BLOCKSIZE_n)) 
			{
				C_s[0] = c[0];
			}
			if (TILE >= (4 * BLOCKSIZE_n)) 
			{
				C_s[2 * BLOCKSIZE_m * BLOCKSIZE_n] = c[1];
			}
			if (TILE >= (8 * BLOCKSIZE_n)) 
			{
				C_s[4 * BLOCKSIZE_m * BLOCKSIZE_n] = c[2];
				C_s[6 * BLOCKSIZE_m * BLOCKSIZE_n] = c[3];
			}
			if (TILE >= (16 * BLOCKSIZE_n)) 
			{
				C_s[8 * BLOCKSIZE_m * BLOCKSIZE_n] = c[4];
				C_s[10 * BLOCKSIZE_m * BLOCKSIZE_n] = c[5];
				C_s[12 * BLOCKSIZE_m * BLOCKSIZE_n] = c[6];
				C_s[14 * BLOCKSIZE_m * BLOCKSIZE_n] = c[7];
			}
		}
		else
		{
			if (TILE >= (2 * BLOCKSIZE_n)) 
			{
				C_s[0] = -c[0];
			}
			if (TILE >= (4 * BLOCKSIZE_n)) 
			{
				C_s[2 * BLOCKSIZE_m * BLOCKSIZE_n] = -c[1];
			}
			if (TILE >= (8 * BLOCKSIZE_n)) 
			{
				C_s[4 * BLOCKSIZE_m * BLOCKSIZE_n] = -c[2];
				C_s[6 * BLOCKSIZE_m * BLOCKSIZE_n] = -c[3];
			}
			if (TILE >= (16 * BLOCKSIZE_n)) 
			{
				C_s[8 * BLOCKSIZE_m * BLOCKSIZE_n] = -c[4];
				C_s[10 * BLOCKSIZE_m * BLOCKSIZE_n] = -c[5];
				C_s[12 * BLOCKSIZE_m * BLOCKSIZE_n] = -c[6];
				C_s[14 * BLOCKSIZE_m * BLOCKSIZE_n] = -c[7];
			}
		}
	}
	__syncthreads();
}

//
// Single Block SGEMM
//
// - Perform matrix-matrix C = A * B for shared memory blocks 
//   of size TILE x TILE, where TILE > 1
//
// Requires BLOCKSIZE_n >= TILE -> enough threads for one tidy per column
//
//

template <int TILE, int BLOCKSIZE_m, int BLOCKSIZE_n, bool MINUS>
__forceinline__ __device__ void block_sgemm3(volatile float *A1_s, volatile float *B1_s, volatile float *C_s, int tidx, int tidy)
{
	float c;
	if (tidx < BLOCKSIZE_m / 2 && tidy < TILE)
	{
		c = 0;
        #pragma unroll
		for (int i = 0; i < TILE; ++i)
		{
			c += A1_s[i * BLOCKSIZE_m] * B1_s[i];
		}
	}
	__syncthreads();
	if (tidx < BLOCKSIZE_m / 2 && tidy < TILE)
	{
		if (!MINUS) C_s[0] = c;
		else C_s[0] = -c;
	}
	__syncthreads();
}


//
// Kernel for computing inverse of diagonal blocks of A in shared memory
// ---------------------------------------------------------------------
// Threads    : BLOCKSIZE_m * BLOCKSIZE_n
// Shared mem : BLOCKSIZE_m * BLOCKSIZE_m
//

template <int BLOCKSIZE_m, int BLOCKSIZE_n, int WORKSIZE_m>
__launch_bounds__(BLOCKSIZE_m * BLOCKSIZE_n, MIN_BLOCKS_PER_MP)
__global__ void strsv_inv_diag_shared(const float* A, int lda, float* x, float* work, int m, int mtile, int ntile)
{
	//printf("VI KOMMER HER: %d %d %d %d %d %d %d\n", BLOCKSIZE_m, BLOCKSIZE_n, WORKSIZE_m, lda, m, work);
	int tidx = threadIdx.x;
	int tidy = threadIdx.y;
	int tid = tidx + tidy * BLOCKSIZE_m;
	int row = blockIdx.x * BLOCKSIZE_m;
	__shared__ float A_smem[BLOCKSIZE_m * BLOCKSIZE_m];
	// read in all data
	A += tidx + tidy * lda + row + row * lda;
	#pragma unroll
	for (int j = 0; j < BLOCKSIZE_m; j += BLOCKSIZE_n)
	{
		A_smem[tid + j * BLOCKSIZE_m] = ((float)(tidx >= j + tidy)) * A[j * lda];
	}
	__syncthreads();
	// do diagonal
	volatile float *A_d = A_smem + tidx + tidx * BLOCKSIZE_m;
	if (tid < BLOCKSIZE_m) A_d[0] = 1.0f / A_d[0];
	__syncthreads();
	// zero out 1st upper off-diagonal
	if (!(tidx & 1)) A_d[BLOCKSIZE_m] = 0;
	// do 1st off-diagonal
	A_d++;
	if (!(tidx & 1)) A_d[0] = -A_d[-1] * A_d[0] * A_d[BLOCKSIZE_m];
	// do 2x2 block
#define TILE 2
	if (BLOCKSIZE_n < TILE)
	{
		volatile float *A_s = A_smem + (tidx - (tidx & (2 * TILE - 1))) * (1 + BLOCKSIZE_m) + TILE + (tid & (TILE - 1));
		volatile float *C_s = A_s + (tidx & TILE) * (BLOCKSIZE_m / TILE);
		volatile float *B_s = C_s - (tid & (TILE - 1));
		block_sgemm2<TILE, BLOCKSIZE_m, BLOCKSIZE_n, false>(A_s, B_s - TILE, C_s, tidx, tid);
		block_sgemm2<TILE, BLOCKSIZE_m, BLOCKSIZE_n, true>(A_s + TILE * BLOCKSIZE_m, B_s, C_s, tidx, tid);
	}
	else
	{
		volatile float *A_s = A_smem + 2 * (tidx - (tidx & (TILE - 1))) * (1 + BLOCKSIZE_m) + TILE + (tidx & (TILE - 1));
		volatile float *C_s = A_s + (tidy * TILE) * (BLOCKSIZE_m / TILE);
		volatile float *B_s = C_s - (tidx & (TILE - 1));
		block_sgemm3<TILE, BLOCKSIZE_m, BLOCKSIZE_n, false>(A_s, B_s - TILE, C_s, tidx, tidy);
		block_sgemm3<TILE, BLOCKSIZE_m, BLOCKSIZE_n, true>(A_s + TILE * BLOCKSIZE_m, B_s, C_s, tidx, tidy);
	}
	if (BLOCKSIZE_m > 4)
	{
		// do 4x4 block
#define TILE 4
		if (BLOCKSIZE_n < TILE)
		{
			volatile float *A_s = A_smem + (tidx - (tidx & (2 * TILE - 1))) * (1 + BLOCKSIZE_m) + TILE + (tidx & (TILE - 1));
			volatile float *C_s = A_s + ((tidx & TILE) + tidy * 2 * TILE) * (BLOCKSIZE_m / TILE);
			volatile float *B_s = C_s - (tidx & (TILE - 1));
			block_sgemm2<TILE, BLOCKSIZE_m, BLOCKSIZE_n, false>(A_s, B_s - TILE, C_s, tidx, tid);
			block_sgemm2<TILE, BLOCKSIZE_m, BLOCKSIZE_n, true>(A_s + TILE * BLOCKSIZE_m, B_s, C_s, tidx, tid);
		}
		else
		{
			volatile float *A_s = A_smem + 2 * (tidx - (tidx & (TILE - 1))) * (1 + BLOCKSIZE_m) + TILE + (tidx & (TILE - 1));
			volatile float *C_s = A_s + (tidy * TILE) * (BLOCKSIZE_m / TILE);
			volatile float *B_s = C_s - (tidx & (TILE - 1));
			block_sgemm3<TILE, BLOCKSIZE_m, BLOCKSIZE_n, false>(A_s, B_s - TILE, C_s, tidx, tidy);
			block_sgemm3<TILE, BLOCKSIZE_m, BLOCKSIZE_n, true>(A_s + TILE * BLOCKSIZE_m, B_s, C_s, tidx, tidy);
		}
	}
	if (BLOCKSIZE_m > 8)
	{
		// do 8x8 block
#define TILE 8
		if (BLOCKSIZE_n < TILE)
		{
			volatile float *A_s = A_smem + (tidx - (tidx & (2 * TILE - 1))) * (1 + BLOCKSIZE_m) + TILE + (tid & (TILE-1));
			volatile float *C_s = A_s + ((tidx & TILE) + tidy * 2 * TILE) * (BLOCKSIZE_m / TILE);
			volatile float *B_s = C_s - (tid & (TILE-1));
			block_sgemm2<TILE, BLOCKSIZE_m, BLOCKSIZE_n, false>(A_s, B_s - TILE, C_s, tidx, tid);
			block_sgemm2<TILE, BLOCKSIZE_m, BLOCKSIZE_n, true>(A_s + TILE * BLOCKSIZE_m, B_s, C_s, tidx, tid);
		}
		else
		{
			volatile float *A_s = A_smem + 2 * (tidx - (tidx & (TILE - 1))) * (1 + BLOCKSIZE_m) + TILE + (tidx & (TILE - 1));
			volatile float *C_s = A_s + (tidy * TILE) * (BLOCKSIZE_m / TILE);
			volatile float *B_s = C_s - (tidx & (TILE - 1));
			block_sgemm3<TILE, BLOCKSIZE_m, BLOCKSIZE_n, false>(A_s, B_s - TILE, C_s, tidx, tidy);
			block_sgemm3<TILE, BLOCKSIZE_m, BLOCKSIZE_n, true>(A_s + TILE * BLOCKSIZE_m, B_s, C_s, tidx, tidy);
		}
	}
	if (BLOCKSIZE_m > 16)
	{
		// do 16x16 block
#define TILE 16
		if (BLOCKSIZE_n < TILE)
		{
			volatile float *A_s = A_smem + (tidx - (tidx & (2 * TILE - 1))) * (1 + BLOCKSIZE_m) + TILE + (tid & (TILE-1));
			volatile float *C_s = A_s + ((tidx & TILE) + tidy * 2 * TILE) * (BLOCKSIZE_m / TILE);
			volatile float *B_s = C_s - (tid & (TILE-1));
			block_sgemm2<TILE, BLOCKSIZE_m, BLOCKSIZE_n, false>(A_s, B_s - TILE, C_s, tidx, tid);
			block_sgemm2<TILE, BLOCKSIZE_m, BLOCKSIZE_n, true>(A_s + TILE * BLOCKSIZE_m, B_s, C_s, tidx, tid);
		}
		else
		{
			volatile float *A_s = A_smem + 2 * (tidx - (tidx & (TILE - 1))) * (1 + BLOCKSIZE_m) + TILE + (tidx & (TILE - 1));
			volatile float *C_s = A_s + (tidy * TILE) * (BLOCKSIZE_m / TILE);
			volatile float *B_s = C_s - (tidx & (TILE - 1));
			block_sgemm3<TILE, BLOCKSIZE_m, BLOCKSIZE_n, false>(A_s, B_s - TILE, C_s, tidx, tidy);
			block_sgemm3<TILE, BLOCKSIZE_m, BLOCKSIZE_n, true>(A_s + TILE * BLOCKSIZE_m, B_s, C_s, tidx, tidy);
		}
	}
		/*
	// solve system as y = A^-1 x
	__shared__ float x_s[BLOCKSIZE_m];
	float *A_s = A_smem + tid;
	x += tidx + row;
	if (tidy == 0) { x_s[tid] = x[0]; x[0] = 0; }
	__syncthreads();
	float sum = 0;
    #pragma unroll
	for (int j = 0; j < BLOCKSIZE_m; j += BLOCKSIZE_n)
	{
		sum += A_s[0] * x_s[j + tidy]; A_s += BLOCKSIZE_m * BLOCKSIZE_n;
	}
	__syncthreads();
	atomicAdd(x, sum);
		*/
	// copy back inverse
	//work += tidx + tidy * lda + row + row * lda; // if at location in A
	work += tidx + tidy * (BLOCKSIZE_m * WORKSIZE_m) + row * (BLOCKSIZE_m * WORKSIZE_m) + (blockIdx.x & (WORKSIZE_m - 1)) * BLOCKSIZE_m;
    #pragma unroll
	for (int j = 0; j < BLOCKSIZE_m; j += BLOCKSIZE_n)
	{
		work[0] = A_smem[tid + j * BLOCKSIZE_m];
		//work += lda * BLOCKSIZE_n; // if at location in A
		work += (BLOCKSIZE_m * WORKSIZE_m) * BLOCKSIZE_n;
	}
}

//
// Kernel for expanding inverse of diagonal blocks of A in device memory
// ---------------------------------------------------------------------
// Threads    : BLOCKSIZE_m
// Shared mem : BLOCKSIZE_m * WORKSIZE_n
//

template <int BLOCKSIZE_m, int WORKSIZE_n, int STEP>
__launch_bounds__(BLOCKSIZE_m, MIN_BLOCKS_PER_MP)
__global__ void strsv_inv_diag_expand(const float *A, int lda, float *B, int ldb, float *C, int tiles, int m, int mtile, int ntile)
{
	int tid = threadIdx.x;
	int row = 2 * blockIdx.x * BLOCKSIZE_m;
	int col = blockIdx.y * WORKSIZE_n;
	int jmp = row * ldb + (blockIdx.x & (tiles - 1)) * (2 * BLOCKSIZE_m);
	__shared__ float B_s[BLOCKSIZE_m * WORKSIZE_n];
	if (STEP == 1) A += tid + row * (1 + lda);
	if (STEP == 2) A += tid + jmp;
	B += tid + col * ldb + jmp;
	C += tid + col * ldb + jmp;

	// Copy columns of B into shared memory
	for (int i = 0; i < WORKSIZE_n; ++i)
	{
		B_s[tid + i * BLOCKSIZE_m] = B[i * ldb];
		// Erase values above diagonal used as workspace in step 1
		if (STEP == 2) B[i * ldb] = 0.0f;
	}
	__syncthreads();

	// Do dot product between row of A and columns of B
	initialize_sums(WORKSIZE_n);
    #pragma unroll
	for (int k = 0; k < BLOCKSIZE_m; ++k)
	{
		if (STEP == 1)
		{
			add_to_sums(WORKSIZE_n, BLOCKSIZE_m); A += lda; // NB: plus in step 1
		}
		else
		{
			sub_to_sums(WORKSIZE_n, BLOCKSIZE_m); A += lda; // NB: minus in step 2

		}
	}
	write_sums(WORKSIZE_n, ldb);
}


